#include <ctype.h>
#include <stdio.h>
#include <stdbool.h>
#include <assert.h>
#include <limits.h>

#define todigit(a) ((a) - '0')

/* convert a string to an int (the base is 10).
 * On success, the pointer at the first non-digit
 * character is returned in endp,
 * on error NULL is stored to endp */
#ifdef __cplusplus
extern "C"
#endif
int strtoi(const char *str, const char **endp)
{
	const char *end;
	bool negative = 0;

	if (*str == '-') {
		negative = 1;
		++str;
	}

	end = str;

	/* move to the end of digits */
	while(*end && isdigit(*end++));

	*endp = end;

	/* end points either to the terminating zero
	 * or to the first non-digit character, so move to
	 * the first digit */
	--end;

	unsigned int pos = 1;
	int result = 0;
	int tmp;
	while (end != str) {
		assert(isdigit(*end));
		tmp = pos * (todigit(*end));
		/* check for overflow */
		assert(tmp >= 0);
		if (INT_MAX > result + tmp) {
			result += tmp;
		} else {
			/* the number is too big to fit to int */
			goto err;
		}

		--end;
		pos *= 10;
		assert(end >= str);
	}

	/* convert the last one */
	assert(end == str);
	assert(pos <= 1000000000 && pos % 10 == 0);

	/* this means numbers >= 3000000000, those do not
	 * fit into an integer */
	if (pos == 1000000000 && todigit(*end > 2))
		goto err;

	tmp = pos * (todigit(*end));
	assert(tmp >= 0);
	if (INT_MAX > tmp - (negative ? 1 : 0)) {
		if (negative && (INT_MAX - tmp == result - 1))
			return INT_MIN;
		else if (INT_MAX - tmp >= result - (negative ? 1 : 0))
			result += tmp;
		else
			goto err;
	} else {
		/* the number is too big to fit to int */
		goto err;
	}

	if (negative)
		return -result;
	else
		return result;

err:
	*endp = NULL;
	return 0;
}

int main(void)
{
	const char* end;

	assert(strtoi("0", &end) == 0);
	assert(strtoi("1", &end) == 1);
	assert(strtoi("-1", &end) == -1);
	assert(strtoi("123", &end) == 123);
	assert(strtoi("-123", &end) == -123 && *end == 0);
	assert(strtoi("134523", &end) == 134523);
	assert(strtoi("-134523", &end) == -134523);
	assert(strtoi("11111", &end) == 11111);
	assert(strtoi("111", &end) == 111 && end && *end == '\0');

	assert(strtoi("assd", &end) && end == NULL);
	assert(strtoi("07423", &end) == 07423);
	assert(strtoi("-2147483648", &end) == -2147483648);
}
